#!/usr/bin/python

#Demonstrate the relative noise in using CDS(m) with multiple reads, vs Proper Linear-regression.
#This shows that Simple Linear regression significantly improves the S/N ratio.

#It should be possible to get a simple proof, but I don't completely trust that... so do it experimentally, in a Monte-Carlo type manner.

#Copyright Richard Neill 2011-2012; this is Free Software Licensed under the GNU GPL version 3+.

#Configurable Parameters:
SAMPLES=100	#Dataset has 100 points in it.
CDS=10		#CDS_multiple picks the first and last x of these to average.  [Must be even, or fix the interpolation of CDS_m's calculation for b]

#Constants.
TRIALS=10000	#Number of times we run the monte-carlo simulation.  10k is good.
A=100		#Offset
B=-1		#Slope
C=10		#Amplitude of (Gaussian) Noise.

#Underlying Physics:
#    CCD pixel (capacitor) is reset to ~ 3V. It is then discharged for about 1ms, by incoming photons, each of which reduce the voltage.
#    Every read has associated thermal noise, because the ADC is imperfect. This noise is larger than the signal...
#    During this time we can sample up to 100 Analog to Digital conversions. Then calculate the gradient (and hence the photon count). Either:
#     (a) LinReg:  we do Linear Regression, using all the samples, according to this model:  http://en.wikipedia.org/wiki/Simple_linear_regression
#     (b) CDSM:    we do Correlated double sampling with multiple reads: read the first 10 points and the last 10 points, average both groups, and then get the gradient.
# Which is better? Clearly LinReg has more information, so should be X times better than CDSM, but what is the value of X?

#Plan:
#1. Generate random data (a slightly downwardly sloping line, with lots of noise.)
#2. Estimate the slope by 2 different methods: CDS(m) or LinReg. (as above).
#3. Repeat steps 1 and 2 lots of times. Compare the standard deviations in their estimators of the slope.
#4. Expect to find that LinReg is a better estimator (lower noise) than CDSm.


#Setup
import sys
from numpy import *
from matplotlib import *
from pylab import *
import scipy
import random
random.seed()


#Help.
def usage():
	print "cdsm-vs-linreg.py: compare the virtues of CDS_m and LinReg using MonteCarlo simulation.\n"
	print "CDSm:    Reading from the Camera is noisy, therefore multiple reads are desirable to reduce read-noise."
	print "         Also, the reset-level is unknown. Thus, correlated-double-sampling, with multiple reads is"
	print "         the established technique. I.e. spend about 10%% of the cycle reading (as many times as possible),"
	print "         then integrate for 80%%, then spend the last 10%% reading (as many times as possible)."
	print "         Then, estimate the photon count by averaging the two sets, and finding the difference."
	print ""
	print "LinReg:  Can we do better? Given that reads are non-destructive, and given that our controller isn't \"busy\""
	print "         during the integration period, why not keep reading? Then use simple least-squares linear-regression"
	print "         to estimate the gradient, (i.e. photon-rate) and hence the photon count."
	print ""
	print "MonteCarlo: How *much* better is LinReg than CDSm? This should be easy to calculate, but not to verify..."
	print "         so, experimentally generate lots of dummy data, and see. This test is currently configured (using "
	print "         parameters at the start of this script), to perform",TRIALS,"trials, each of which uses", SAMPLES
	print "         data-points, of which",CDS,"are used at start/end for CDS_m."
	print ""
	print "Results: Here are some results from running this program: LinReg always improves S/N by some factor over CDSm:"
	print "               TRIALS     SAMPLES     CDSm         LinReg_S/N_Improvement"
	print "               10k        100         10           1.42"
	print "               10k        100         50           1.15"
	print "               10k        100          2           2.92"
	print "               10k        1000       100           1.44"
	print "               10k        1000        10           4.08"
	print "               100k       100         10           1.43"
	print "         => improvement is approximately linearly proportional to (SAMPLES/CDSm)."
	print ""
	#Plot graph?  For the data above, plot y=SN_improvement, x=(SAMPLES/CDSm). It's fairly linear.
	#plt.figure()
	#plt.plot ([2,10,50,100], [1.15, 1.42, 2.92, 4.08])
	#plt.show()
	return


#Trivial arg parsing.
if (len(sys.argv) > 1):
	usage()
	exit(1)


#Monte-Carlo: do the following multiple times:
linreg = []
cdsm = []
estimate_s = 20*TRIALS/10000  #by experiment on 1.8 GHz atom cpu.
print "Now running",TRIALS,"monte-carlo-trials: this may take a few seconds (estimate:",estimate_s,"s)..."
for trial in range(TRIALS):


	#Generate some data:  y = A + Bx + C.noise()
	x = arange (0.0, SAMPLES)
	noise = []
	for i in range(SAMPLES): noise.append(random.gauss(0,1))
	noise = array(noise)

	y = A +   B * x   + C * noise


	#Do simple linear regression, least-squares. See linregplot.py for explanation of formulae.
	n       = len(x)
	S_x 	= sum(x)
	S_xx    = sum(x * x)
	S_y     = sum(y)
	S_yy    = sum(y * y)
	S_xy    = sum(x * y)

	b       = ( n * S_xy -  S_x * S_y ) /  ( n * S_xx - (S_x * S_x) )	#b-hat, estimator for gradient.
	a       = ( S_y / n ) - ( b * S_x / n)					#a-hat, estimator for y-intercept.
	s = sqrt(abs( (1.0 / (n * (n-2))) * ( n * S_yy - (S_y * S_y) - (b * b * (n * S_xx - (S_x * S_x)) ) )))	#sigma-hat, (estimator of std-dev of noise)
	se_b =  sqrt (abs ( ( n * (s * s) ) / ( n * S_xx - (S_x * S_x) ) ) )	#std err in b-hat
	se_a = sqrt ( (se_b * se_b) * S_xx / n )				#std err in a-hat
	r = ( (n * S_xy) - (S_x * S_y) ) / sqrt(abs( (n * S_xx - (S_x * S_x)) * (n * S_yy - (S_y * S_y)) ))	#r-hat, estimator for Pearson's product-moment-correlation-coefficient.
	delta_y = y[n-1] - y[0]							#absolute change in y
	b_delta_x = b * (x[n-1] - x[0])						#predicted change in y.
	#print "DebugLR: S_x=", S_x, " S_y=", S_y, " S_xx=", S_xx, " S_yy=", S_yy, " S_xy=", S_xy, " n=",  n, " a=",  a, " b=",  b, " s=",  s, " se_a=",  se_a, " se_b=",  se_b, " r=",  r, " delta-y=",  delta_y, " b.delta-x=",  b_delta_x

	linreg_estimate = b_delta_x	#LinReg's "answer"
	linreg.append(linreg_estimate)

	#Now do CDS_m calculation of gradient.
	first = sum(y[0:CDS])							#first 10
	last = sum(y[-CDS:]) 							#last 10
	b = ((last - first) / CDS) / (x[SAMPLES-(CDS/2)] - x[(CDS/2)]) 		#average_of_last - average_of_first / run-length,  where run_length is from CDS/2  to  (SAMPLES - CDS/2)
	delta_y = y[n-1] - y[0]							#absolute change in y
	b_delta_x = b * (x[n-1] - x[0])						#predicted change in y.
	#print "DebugCDSm: first=", first, " last =", last, " b=", b, " n=", n, " m=", CDS, " delta-y=",  delta_y, " b.delta-x=",  b_delta_x

	cdsm_estimate = b_delta_x	#CDS_m's "answer"
	cdsm.append(cdsm_estimate)

	#Compare:
	#print "Underlying Delta_Y is: ", B*(SAMPLES-1), ". CDSm estimate is: ", cdsm_estimate, ". Linear Regression estimate is: ", linreg_estimate


#Which is better?
linreg_mean = mean(linreg)
linreg_stddev = std(linreg, ddof=1)

cdsm_mean = mean(cdsm)
cdsm_stddev = std(cdsm, ddof=1)

linreg_stddev_win = cdsm_stddev / linreg_stddev
diff_means = linreg_mean - cdsm_mean


#Summarise
print "Completed", TRIALS, "Monte-Carlo trials. Data-set has", SAMPLES, "samples, of which", CDS, "are used at each end for CDS_m, and all", SAMPLES, "are used for linear regression."
print "Underlying data is \"y = A +  B * x  + C * noise\", where A=", A, "B=",B, "C=",C,". Delta-Y over ",SAMPLES,"samples should be ",B*(SAMPLES-1),"."
print "Linear regression M-C gives obtains value of Delta-Y with (mean,stddev) =", (linreg_mean, linreg_stddev)
print "CDS_multiple M-C gives value of Delta-Y with (mean,stddev) =", (cdsm_mean, cdsm_stddev)
print "Difference in means is: ", diff_means, ".  <- this should be approximately zero, given that both LinReg and CDSm are unbiased."
print "Difference in StdDevs is a factor of ", linreg_stddev_win ," i.e. Using LinReg Improves the S/N ratio by a factor of ", linreg_stddev_win, " compared to CDS_m."
print ""

