#!/usr/bin/python

#This program reads a data-file, and outputs a WAV file. It can also apply gain.
#If the file has several columns, we can choose any one or the sum, or all of them.

#Copyright (C) Richard Neill 2011-2012, <maths at REMOVE.ME.richardneill.org>. This program is Free Software. You can
#redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
#either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
#For the details, please see: http://www.gnu.org/licenses/gpl.html

#Thanks to Mike Rudin for pointing out that the human ear is much more sensitive than an FFT plot for distinguishing noise.

#Begin
from __future__ import division		#Needed, as this may run under Python 2.x
import getopt, sys, os
import numpy as numpy
from pylab import *
from scipy.io import wavfile


#Constants
MAXVAL = (2**31 - 1)  			#Max val for a signed 32 bit int, used in wav file.
NORMALISE = MAXVAL / sqrt(2)		#Normalise target.

#Help
def usage():
	print 	"dat2wav:  read in a data-file, and convert it to a 32-bit WAV file.\n"
	print 	"USAGE:    " + str(os.path.basename(sys.argv[0])) + "  [OPTS]  infile.dat  outfile.wav\n"
	print	"OPTIONS:  -f, --samplefreq  FREQ         define the (integer) sampling frequency in Hz. [default: 10,000]."
	print	"          -g, --gain        G,normalise  apply a signal gain of G, (or normalise). [default: normalise]."
	print   "          -c, --column      N,sum,all    use Nth column's data (N is 0-based). [default: all]."
	print   "          -d, --keepdc                   keep the DC offset. [default: subtract offset]."
	print   "          -l, --loop        N            loop N times, concatenating result. [default: 1]."
	print   "          -x, --overwrite                allow overwriting an existing output file [default: no]."
	print   "          -h, --help                     show help.\n"
	print   "INPUT:    Input file, infile.dat ('-' for stdin) should be columns of (int/float) data, separated by"
	print   "          whitespace, with optional #comments, i.e. the standard form for numpy.loadfile(). cf. genffttest."
	print   "          (compressed files (.gz/.bz2) can also be read transparently, no need to uncompress first.)\n"
	print   "NOTES:    * This is useful to view the output in a wav editor, (e.g. 'audacity') or hear it (e.g. 'play')."
	print   "          * It's also much easier to *hear* certain types of noise than to see it in an FFT spectrum."
	print   "          * The sample frequency should be chosen to optimise listening, not necessarily the original value."
	print   "          * Normalisation sets the peak to 0.707, rather than to 1, so as to avoid distortion during playback."
	print   "          * To see the raw data, use '-d -g 1'."
	print   ""
	return

#Defaults
samplerate = 10000
column = -1
gain = 1
overwrite = loop = False
normalise = removedc = True


#Parse args
try:
	opts, args = getopt.gnu_getopt(sys.argv[1:], "dc:f:g:hl:x",
	  ["keepdc", "column=", "samplefreq=", "gain=", "help", "loop=", "overwrite"])
except getopt.GetoptError, err:
	print str(err) + ": use -h for help." ; sys.exit(2)

for o, a in opts:
	if o in ("-h", "--help"):
		usage(); exit(0);
	elif o in ("-f", "--samplefreq"):
		samplerate = int(a)
	elif o in ("-d", "--keepdc"):
		removedc = False
	elif o in ("-c", "--column"):
		if (a == "all"):
			column = a
		elif (a == "sum"):
			column = a
		else:
			column = int(a)
	elif o in ("-g", "--gain"):
		if (a == "normalise"):
			gain = 0
		else:
			normalise = False
			gain = float(a)
	elif o in ("-x", "--overwrite"):
		overwrite = True
	elif o in ("-l", "--loop"):
		loop = int(a)
	else:
		assert False, "unhandled option"

#NB, support "-" for stdin, but can't support "-" for stdout, unless we alter all the print() statements to use stderr.
if (len(args) == 2):
	infile = args[0];
	outfile = args[1];
	infilename = infile
	if (infile == "-"):
		infile = sys.stdin
		infilename = "<stdin>"
elif ( (len(args) == 1) and (sys.argv.count('-') == 1) ):	#Support older versions too.
	infile = sys.stdin
	outfile = args[0]
	infilename = "<stdin>"
else:
	print "Error: Two (non-option) arguments are required: input-filename.dat, output-filename.wav."
	exit (2)

if (os.path.exists(outfile) and (overwrite == False)):
	print "Error: output file '" + outfile + "' already exists. Use -x to allow overwrite."
	exit (2)


#Load data from file.
try:
	data_in = numpy.loadtxt (infile);				#numpy.loadtxt really is all we need! It even handles comments :-)
except IOError, err:							#the function is very flexible, though the defaults suffice here.
	print "Error reading datafile " + infilename + ": " + str(err)	#data_in is an ndarray of floats.
	exit (1);
except ValueError, err:
	print "Error processing malformed datafile " + infilename + ": " + str(err)
	exit (1);


#Get size of data file, sanity check it.
if (len(data_in.shape) == 1):		#if it is a single column (1d-array),  convert to a 2d-array with 1 column.
	data_in = data_in[:,newaxis]
num_cols = len(data_in[0, ]);
num_rows = len(data_in);
if (num_cols == 0) or (num_cols > 20) or (num_rows < 4) or (num_rows > 1E7):
	print "Data size is out of range. File contains "  + str(num_rows) + " rows of " + str(num_cols) + " columns."
	exit (1)


#Extract the relevant column.
if (column == -1):			#Default: All columns.
	column = "all"
	if (num_cols == 1):		#..but if the file only has 1 column, don't make it an array (for clarity).
		column = 0
if (column == "sum"):			#Sum the columns
	data = data_in [ : , 0]
	for col in range (1, num_cols):
		data += data_in [ : , col]
	num_channels = 1
elif (column == "all"):			#All columns (in parallel).
	data = data_in
	num_channels = num_cols
elif (column < num_cols):
	data = data_in [ : , column]	#This pulls out a 1-d array, from a 2-d array.
	num_channels = 1
else:
	assert False, "invalid column specified: " + str(column)


#Remove DC offset by subtracting mean.
if (removedc):
	mean = data.mean(axis=0)
	data = data - mean;
	dc_str = "subtracted"
else:
	dc_str = "unaltered"


#Normalise:calculate gain required that will make exactly one value hit the precise limit.
if (normalise):
	max = data.max(axis=0)
	maxneg = abs (data.min(axis=0))
	peak = maximum (max,maxneg)
	gain = NORMALISE / peak
	gain_str = "normalise; " +str(gain.round(decimals=2))
else:
	gain_str = str(round(gain,2))


#Apply gain
data = data * gain;


#Detect clipping
if ( (data.max() > MAXVAL) or (data.min() < - MAXVAL) ):
	print "\nWARNING: Clipping has occurred. Data should be within range ["+str(-MAXVAL)+", "+str(MAXVAL)+"] but is actually ["+str(data.min())+", "+str(data.max())+"].\n"
	clip_str = "YES!"
else:
	clip_str = "no"


#Loop (concatenate wave to self, n times)
if (loop):
	data2 = data
	for i in range (1, loop):
		data2 = append(data2, data, axis=0);
	data = data2
	loop_str = str(loop)
else:
	loop_str = "no"


#Output data as a wavfile. ('scipy.io.wavfile' is much simpler than 'wave')
data = data.astype(int32)  #Sets the bits-per-sample to 32 for the wavfile.
num_samples = len(data);
seconds = (num_samples * 1.0) / samplerate
wavfile.write(outfile, samplerate, data)

#Summarise.
print "Input file: " + infilename + ". Column: " +str(column)+ ". Gain: " + gain_str + ". DC Offset: " + dc_str + ". Loops: " + loop_str + ". Clipped: " + clip_str + "."
print "Output file: "+ outfile +". 32 bit. "+ str(num_channels) +" channels. "+ str(samplerate) +" Hz. "+ str(num_samples) +" Samples ("+ str(seconds) +" seconds)."
exit(0)
